'use strict';

// ensure it works with frozen one

Object.freeze(Object.prototype);

var chai = require('chai');

//
var $as = typeof window !== 'undefined' ? require('futoin-asyncsteps') : module.require('../lib/main-full');

var assert = chai.assert,
    expect = chai.expect;
var Mutex = $as.Mutex;


describe('Mutex', function () {
    it('should handle re-entrancy & success params', function (done) {
        var mtx = new Mutex();

        var _loop = function _loop(i) {
            var as = $as();

            as.add(function (as) {
                return as.success('MyArgs');
            });
            as.sync(mtx, function (as, arg) {
                expect(arg).to.equal('MyArgs');

                as.sync(mtx, function (as) {
                    return as.error('Wrong');
                }, function (as, err) {
                    if (err === 'Wrong') {
                        as.error('OK');
                    }
                });
            }, function (as, err) {
                as.success(err);
            });
            as.add(function (as, err) {
                if (err === 'OK') {
                    if (i) {
                        done();
                    }
                } else {
                    done(as.state.last_exception || 'Fail');
                }
            });
            as.execute();
        };

        for (var i = 0; i < 2; ++i) {
            _loop(i);
        }
    });

    it('should correctly handle concurrency limit', function (done) {
        var limit = 3;
        var mtx = new Mutex(limit);
        var curr = 0;
        var max = 0;

        for (var i = limit * 3; i >= 0; --i) {
            var as = $as();
            as.sync(mtx, function (as) {
                curr += 1;
                max = max < curr ? curr : max;

                as.add(function (as) {
                    // break burst execution
                    as.waitExternal();
                    setTimeout(function () {
                        curr -= 1;
                        as.state && as.success();
                    }, 0);
                });
            });

            if (i === 0) {
                as.add(function (as) {
                    try {
                        expect(max).to.equal(limit);
                        done();
                    } catch (e) {
                        done(e);
                    }
                });
            }

            as.execute();
        }
    });

    it('should handle cancel', function (done) {
        var mtx = new Mutex();
        var as1 = $as();

        as1.add(function (as) {
            as.sync(mtx, function (as) {
                as.waitExternal();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();

        $as().add(function (as) {
            as.setTimeout(10);
            as.sync(mtx, function (as) {
                as.waitExternal();
            });
        }, function (as, err) {
            if (err === 'Timeout') {
                as1.cancel();
            } else {
                done(as.state.last_exception || 'Fail');
            }
        }).execute();

        $as().add(function (as) {
            as.sync(mtx, function (as) {
                done();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();
    });

    it('should handle errors', function (done) {
        var mtx = new Mutex();
        var as1 = $as();
        var as1p = void 0;

        as1.add(function (as) {
            as.sync(mtx, function (as) {
                as.waitExternal();
                as1p = as;
            });
        }, function (as, err) {
            if (err === 'OK') {
                as.success();
            } else {
                done(as.state.last_exception || 'Fail');
            }
        }).execute();

        $as().add(function (as) {
            as.loop(function (as) {
                if (as1p) {
                    try {
                        as1p.error('OK');
                    } catch (e) {
                        // ignore
                    }

                    as.break();
                }
            });
            as.sync(mtx, function (as) {
                done();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();
    });

    it('should handle parallel steps', function (done) {
        var mtx = new Mutex();
        var curr = 0;
        var max = 0;
        var as = $as();
        var p = as.parallel();

        for (var i = 10; i >= 0; --i) {
            p.add(function (as) {
                as.sync(mtx, function (as) {
                    curr += 1;
                    max = max < curr ? curr : max;

                    as.add(function (as) {
                        curr -= 1;
                    });
                });
            });

            if (i === 0) {
                as.add(function (as) {
                    try {
                        expect(max).to.equal(1);
                        done();
                    } catch (e) {
                        done(e);
                    }
                });
            }
        }

        as.execute();
    });

    it('should handle queue limit', function (done) {
        var mtx = new Mutex(2, 5);
        var curr = 0;
        var max = 0;
        var total = 0;
        var rejected = 0;
        var as = $as();
        var p = as.parallel();

        for (var i = 10; i > 0; --i) {
            p.add(function (as) {
                as.add(function (as) {
                    return as.sync(mtx, function (as) {
                        curr += 1;
                        total += 1;
                        max = max < curr ? curr : max;

                        as.add(function (as) {
                            as.waitExternal();
                            setTimeout(function () {
                                return as.state && as.success();
                            }, 100);
                        });
                        as.add(function (as) {
                            curr -= 1;
                        });
                    });
                }, function (as, err) {
                    if (err === 'DefenseRejected') {
                        rejected += 1;
                        as.success();
                    }
                });
            });
        }

        as.add(function (as) {
            try {
                expect(max).to.equal(2);
                expect(total).to.equal(7);
                expect(rejected).to.equal(3);
                done();
            } catch (e) {
                done(e);
            }
        });

        as.execute();
    });
});
//# sourceMappingURL=mutextest.js.map
