'use strict';

// ensure it works with frozen one

Object.freeze(Object.prototype);

var chai = require('chai');

//
var $as = typeof window !== 'undefined' ? require('futoin-asyncsteps') : module.require('../lib/main-full');

var assert = chai.assert,
    expect = chai.expect;
var Throttle = $as.Throttle;

var _require = require('../lib/common'),
    TIMER = _require.TIMER;

describe('Throttle', function () {
    it('should handle re-entrancy & success params', function (done) {
        var thrtl = new Throttle(2, 10);

        var _loop = function _loop(i) {
            var as = $as();

            as.add(function (as) {
                return as.success('MyArgs');
            });
            as.sync(thrtl, function (as, arg) {
                expect(arg).to.equal('MyArgs');

                as.sync(thrtl, function (as) {
                    return as.error('Wrong');
                }, function (as, err) {
                    if (err === 'Wrong') {
                        as.error('OK');
                    }
                });
            }, function (as, err) {
                as.success(err);
            });
            as.add(function (as, err) {
                if (err === 'OK') {
                    if (i) {
                        done();
                    }
                } else {
                    done('Fail');
                }
            });
            as.execute();
        };

        for (var i = 0; i < 2; ++i) {
            _loop(i);
        }
    });

    it('should correctly handle concurrency limit', function (done) {
        var limit = 3;
        var thrtl = new Throttle(limit, 100);
        var curr = 0;
        var max = 0;

        for (var i = limit * 3; i >= 0; --i) {
            var as = $as();
            as.sync(thrtl, function (as) {
                curr += 1;
                max = max < curr ? curr : max;

                as.add(function (as) {
                    // break burst execution
                    as.waitExternal();
                    $as.ActiveAsyncTool.callLater(function () {
                        curr -= 1;
                        as.success();
                    }, 0);
                });
            });

            if (i === 0) {
                as.add(function (as) {
                    try {
                        expect(max).to.equal(limit);
                        done();
                    } catch (e) {
                        done(e);
                    }
                });
            }

            as.execute();
        }
    });

    it('should handle cancel', function (done) {
        var thrtl = new Throttle(1, 20);
        var as1 = $as();

        as1.add(function (as) {
            as.sync(thrtl, function (as) {
                as.waitExternal();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();

        $as().add(function (as) {
            as.setTimeout(10);
            as.sync(thrtl, function (as) {
                as.waitExternal();
            });
        }, function (as, err) {
            if (err === 'Timeout') {
                as1.cancel();
            } else {
                done(as.state.last_exception || 'Fail');
            }
        }).execute();

        $as().add(function (as) {
            as.sync(thrtl, function (as) {
                done();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();
    });

    it('should handle errors', function (done) {
        var thrtl = new Throttle(1, 10);
        var as1 = $as();
        var as1p = void 0;

        as1.add(function (as) {
            as.sync(thrtl, function (as) {
                as.waitExternal();
                as1p = as;
            });
        }, function (as, err) {
            if (err === 'OK') {
                as.success();
            } else {
                done(as.state.last_exception || 'Fail');
            }
        }).execute();

        $as().add(function (as) {
            as.loop(function (as) {
                if (as1p) {
                    try {
                        as1p.error('OK');
                    } catch (e) {
                        // ignore
                    }

                    as.break();
                }
            });
            as.sync(thrtl, function (as) {
                done();
            });
        }, function (as, err) {
            done(as.state.last_exception || 'Fail');
        }).execute();
    });

    it('should handle parallel steps', function (done) {
        var thrtl = new Throttle(1, 10);
        var curr = 0;
        var max = 0;
        var as = $as();
        var p = as.parallel();

        for (var i = 3; i >= 0; --i) {
            p.add(function (as) {
                as.sync(thrtl, function (as) {
                    curr += 1;
                    max = max < curr ? curr : max;

                    as.add(function (as) {
                        curr -= 1;
                    });
                });
            });

            if (i === 0) {
                as.add(function (as) {
                    try {
                        expect(max).to.equal(1);
                        done();
                    } catch (e) {
                        done(e);
                    }
                });
            }
        }

        as.execute();
    });

    it('should have correct average throughput', function (done) {
        var thrtl = new Throttle(10, 100);
        var as = $as();
        var p = as.parallel(function (as, err) {
            done(as.state.last_exception || err || 'Fail');
        });
        var passed = 0;

        for (var i = 0; i < 100; ++i) {
            p.add(function (as) {
                as.sync(thrtl, function (as) {
                    // break burst execution
                    as.waitExternal();
                    $as.ActiveAsyncTool.callLater(function () {
                        passed += 1;
                        as.success();
                    }, 0);
                });
            });
        }

        as.execute();

        $as.ActiveAsyncTool.callLater(function () {
            expect(passed).to.equal(50);
            as.cancel();
            done();
        }, 450);
    });

    it('should handle queue limit', function (done) {
        var thrtl = new Throttle(10, 100, 33);
        var as = $as();
        var p = as.parallel(function (as, err) {
            done(as.state.last_exception || err || 'Fail');
        });
        var passed = 0;
        var rejected = 0;

        for (var i = 0; i < 100; ++i) {
            p.add(function (as) {
                as.add(function (as) {
                    return as.sync(thrtl, function (as) {
                        passed += 1;
                    });
                }, function (as, err) {
                    if (err === 'DefenseRejected') {
                        rejected += 1;
                        as.success();
                    }
                });
            });
        }

        as.execute();

        $as.ActiveAsyncTool.callLater(function () {
            expect(passed).to.equal(43);
            expect(rejected).to.equal(57);
            as.cancel();
            done();
        }, 450);
    });

    it('should must stop and resume reset timer', function (done) {
        var thrtl = new Throttle(2, 50, 10);

        $as().add(function (outer_as) {
            var as = $as();
            var p = as.parallel(function (as, err) {
                done(as.state.last_exception || err || 'Fail');
            });
            var passed = 0;
            var rejected = 0;

            for (var i = 0; i < 3; ++i) {
                p.add(function (as) {
                    as.add(function (as) {
                        return as.sync(thrtl, function (as) {
                            passed += 1;
                        });
                    }, function (as, err) {
                        if (err === 'DefenseRejected') {
                            rejected += 1;
                            as.success();
                        }
                    });
                });
            }

            as.execute();

            outer_as.waitExternal();
            $as.ActiveAsyncTool.callLater(function () {
                expect(passed).to.equal(3);
                expect(rejected).to.equal(0);
                as.cancel();
                outer_as.success();
            }, 75);
        }).add(function (outer_as) {
            outer_as.waitExternal();
            $as.ActiveAsyncTool.callLater(function () {
                expect(thrtl._timer).to.be.null;
                outer_as.success();
            }, 100);
        }).add(function (outer_as) {
            var as = $as();
            var p = as.parallel(function (as, err) {
                done(as.state.last_exception || err || 'Fail');
            });
            var passed = 0;
            var rejected = 0;

            for (var i = 0; i < 3; ++i) {
                p.add(function (as) {
                    as.add(function (as) {
                        return as.sync(thrtl, function (as) {
                            passed += 1;
                        });
                    }, function (as, err) {
                        if (err === 'DefenseRejected') {
                            rejected += 1;
                            as.success();
                        }
                    });
                });
            }

            as.execute();

            outer_as.waitExternal();
            $as.ActiveAsyncTool.callLater(function () {
                expect(passed).to.equal(3);
                expect(rejected).to.equal(0);
                as.cancel();
                outer_as.success();
            }, 75);
        }).add(function (as) {
            return done();
        }).execute();
    });
});
//# sourceMappingURL=throttletest.js.map
